from __future__ import annotations
import string
from collections import OrderedDict

import toml

NUMERIC_TYPE_SIZES: dict[str, int] = {
    "byte":   1,
    "bool":   1,
    "uint16": 2,
    "int16":  2,
    "uint32": 4,
    "int32":  4,
    "uint64": 8,
    "int64":  8,
    "float":  4,
    "double": 8,
}

NUMERIC_TYPE_RANGES: dict[str, tuple[int,int]] = {
    "byte":   (                         0,                        255),
    "int16":  (                   -32_768,                     32_767),
    "uint16": (                         0,                     65_535),
    "int32":  (            -2_147_483_648,              2_147_483_647),
    "uint32": (                         0,              4_294_967_295),
    "int64":  (-9_223_372_036_854_775_808,  9_223_372_036_854_775_807),
    "uint64": (                         0, 18_446_744_073_709_551_615),
}


RESERVED_WORDS: list[str] = list(NUMERIC_TYPE_SIZES.keys()) + ["string", "list"]

def _contains_whitespace(s: str) -> bool:
    return True in [c in s for c in string.whitespace]

class Variable():
    def __init__(self, protocol: Protocol, name: str, vartype: str, is_list: bool = False):
        self.protocol = protocol
        self.name = name
        self.vartype = vartype
        self.is_list = is_list

    def __str__(self) -> str:
        return f"Variable: <{self.name} ({'list of ' if self.is_list else ''}{self.vartype})>"

    def __repr__(self) -> str:
        return self.__str__()

    def is_simple(self, listed: bool = False) -> bool:
        if self.is_list and not listed:
            return False
        elif self.vartype in NUMERIC_TYPE_SIZES:
            return True
        elif self.vartype == "string":
            return False
        elif self.vartype in self.protocol.structs:
            return self.protocol.structs[self.vartype].is_simple()
        elif self.vartype in self.protocol.messages:
            return self.protocol.messages[self.vartype].is_simple()
        else:
            raise NotImplementedError(f"Can't determine simplicity of {self.name}.")

class Struct():
    def __init__(self, name: str):
        self.name = name
        self.members: list[Variable] = []
        self.is_message: bool = False

    def is_simple(self) -> bool:
        return all([var.is_simple() for var in self.members])

class Protocol():
    def __init__(self, filename: str = None):
        self.namespace: str = None
        self.list_size_type: str = "uint32"
        self.string_size_type: str = "uint32"
        self.structs: OrderedDict[str,Struct] = OrderedDict()
        self.messages: OrderedDict[str,Struct] = OrderedDict()

        if filename == None:
            return

        self.protocol_string = open(filename, "r", encoding="utf-8").read()
        protocol_data = toml.loads(self.protocol_string)

        if "meta" in protocol_data:
            if "namespace" in protocol_data["meta"]:
                if _contains_whitespace(protocol_data["meta"]["namespace"]):
                    raise ValueError(f"Namespace cannot contain whitespace: '{protocol_data['meta']['namespace']}'")
                self.namespace = protocol_data["meta"]["namespace"]
            valid_sizes = ["byte", "uint16", "int16", "uint32", "int32", "uint64", "int64"]
            if "list_size_type" in protocol_data["meta"]:
                if protocol_data["meta"]["list_size_type"] not in valid_sizes:
                    raise ValueError(f"List size type is not valid: '{protocol_data['meta']['list_size_type']}'")
                self.list_size_type = protocol_data["meta"]["list_size_type"]
            if "string_size_type" in protocol_data["meta"]:
                if protocol_data["meta"]["string_size_type"] not in valid_sizes:
                    raise ValueError(f"String size type is not valid: '{protocol_data['meta']['string_size_type']}'")
                self.string_size_type = protocol_data["meta"]["string_size_type"]

        if "structs" not in protocol_data: protocol_data["structs"] = []
        if "messages" not in protocol_data: protocol_data["messages"] = []

        def build_struct(data: dict[str,any], is_message: bool = False):
            if "_name" not in data:
                raise ValueError(f"Missing _name on {data}")
            s = Struct(data["_name"])
            s.is_message = is_message
            for var_name, var_type in data.items():
                if var_name.startswith("_"): continue
                if var_type[0] == "[" and var_type[-1] == "]":
                    listed_type = var_type[1:-1]
                    if listed_type[0] == "[" and listed_type[-1] == "]":
                        raise ValueError(f"Cannot have list of lists")
                    var = Variable(self, var_name, listed_type, True)
                else:
                    var = Variable(self, var_name, var_type, False)
                s.members.append(var)
            return s

        for struct_data in protocol_data["structs"]:
            s = build_struct(struct_data)
            if s.name in self.structs:
                raise ValueError(f"Duplicate _name on {struct_data}")
            self.structs[s.name] = s
        for message_data in protocol_data["messages"]:
            m = build_struct(message_data, True)
            if m.name in self.messages:
                raise ValueError(f"Duplicate _name on {message_data}")
            self.messages[m.name] = m

        def validate_struct(s: Struct):
            label = "Struct"
            if s.is_message:
                label = "Message"
            if _contains_whitespace(s.name):
                raise ValueError(f"{label} _name cannot contain whitespace: '{s.name}'")
            if s.name in RESERVED_WORDS:
                raise ValueError(f"{label} _name is reserved word: '{s.name}'")
            if s.is_message and s.name in self.structs:
                raise ValueError(f"Message name cannot shadow struct: '{s.name}'")
            for var in s.members:
                if _contains_whitespace(var.name):
                    raise ValueError(f"Member name cannot contain whitespace: '{var.name}'")
                valid_types: list[str] = list(self.structs) + list(NUMERIC_TYPE_SIZES.keys()) + ["string"]
                if var.vartype not in valid_types:
                    if s.is_message and var.vartype in self.messages:
                        raise ValueError(f"Messages cannot contain other messages ({s.name} contains {var.vartype})")
                    raise NotImplementedError(f"No type called {var.vartype} (definition: {s.name})")

        [validate_struct(s) for s in self.structs.values()]
        [validate_struct(m) for m in self.messages.values()]

        if len(self.messages) > 255:
            raise ValueError("Cannot, at present, have more than 255 types of messages. Sorry. :(")

        def check_for_cycles(s: Struct):
            encountered = set([s.name])
            frontier = [m.vartype for m in s.members if m.vartype in self.structs]
            while len(frontier) > 0:
                new_type = frontier.pop()
                if new_type in encountered:
                    raise RecursionError(f"{s.name} and {new_type} reference each other")
                for cvar in self.structs[new_type].members:
                    if cvar.vartype in encountered:
                        raise RecursionError(f"{s.name} and {cvar.vartype} reference each other")
                    if cvar.vartype not in self.structs: continue
                    frontier.append(cvar.vartype)

        [check_for_cycles(s) for s in self.structs.values()]
        [check_for_cycles(m) for m in self.messages.values()]

    def __str__(self) -> str:
        return f"Protocol (namespace: {self.namespace}, {len(self.structs)} structs, {len(self.messages)} messages)"

    def get_size_of(self, var_type: str) -> int:
        if var_type in NUMERIC_TYPE_SIZES:
            return NUMERIC_TYPE_SIZES[var_type]
        elif var_type == "string":
            raise NotImplementedError(f"Cannot pre-calculate size of string")
        else:
            size = 0
            if var_type in self.structs:
                st = self.structs[var_type]
            else:
                st = self.messages[var_type]
            for var in st.members:
                if var.is_list:
                    raise NotImplementedError(f"Cannot calculate size of struct with list ({var_type})")
                size += self.get_size_of(var.vartype)
            return size



from ..protocol import Protocol, Variable, Struct, NUMERIC_TYPE_SIZES
from ..writer import Writer
from .. import LIB_NAME, LIB_VERSION

LANGUAGE_NAME = "Swift"


class SwiftWriter(Writer):
    language_name = LANGUAGE_NAME
    default_extension = ".swift"

    def __init__(self, p: Protocol, extra_args: dict[str,any] = {}):
        super().__init__(protocol=p, tab="    ")

        self.embed_protocol = extra_args["embed_protocol"]

        self.type_mapping["byte"] = "UInt8"
        self.type_mapping["bool"] = "Bool"
        self.type_mapping["uint16"] = "UInt16"
        self.type_mapping["int16"] = "Int16"
        self.type_mapping["uint32"] = "UInt32"
        self.type_mapping["int32"] = "Int32"
        self.type_mapping["uint64"] = "UInt64"
        self.type_mapping["int64"] = "Int64"
        self.type_mapping["float"] = "Float32"
        self.type_mapping["double"] = "Float64"
        self.type_mapping["string"] = "String"

        self.base_defaults: dict[str,str] = {
            "byte": "0",
            "bool": "false",
            "uint16": "0",
            "int16": "0",
            "uint32": "0",
            "int32": "0",
            "uint64": "0",
            "int64": "0",
            "float": "0.0",
            "double": "0.0",
            "string": '""',
        }

    def deserializer(self, var: Variable, accessor: str):
        if var.is_list:
            self.write_line(f"let {var.name}_Length = try dataReader.Get{self.get_native_list_size()}()")
            self.write_line(f"{accessor}{var.name} = []")
            self.write_line(f"for _ in 0..<{var.name}_Length {{")
            self.indent_level += 1
            inner = Variable(self.protocol, f"let {var.name}_el", var.vartype)
            self.deserializer(inner, "")
            self.write_line(f"{accessor}{var.name}.append({var.name}_el)")
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype in NUMERIC_TYPE_SIZES or var.vartype == "string":
            self.write_line(f"{accessor}{var.name} = try dataReader.Get{self.type_mapping[var.vartype]}()")
        else:
            self.write_line(f"{accessor}{var.name} = try {var.vartype}.FromBytes(dataReader: dataReader)")

    def serializer(self, var: Variable, accessor: str):
        if var.is_list:
            self.write_line(f"dataWriter.Write{self.get_native_list_size()}({self.get_native_list_size()}({accessor}{var.name}.count))")
            self.write_line(f"for el in {accessor}{var.name} {{")
            self.indent_level += 1
            inner = Variable(self.protocol, "el", var.vartype)
            self.serializer(inner, "")
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype in NUMERIC_TYPE_SIZES or var.vartype == "string":
            self.write_line(f"dataWriter.Write{self.type_mapping[var.vartype]}({accessor}{var.name})")
        else:
            self.write_line(f"{accessor}{var.name}.WriteBytes(dataWriter)")

    def gen_measurement(self, st: Struct, accessor: str = "") -> tuple[list[str], int]:
        lines: list[str] = []
        accum = 0

        if st.is_simple():
            lines.append(f"return {self.protocol.get_size_of(st.name)};")
        else:
            size_init = "var size = 0"
            lines.append(size_init)

            for var in st.members:
                if var.is_list:
                    accum += NUMERIC_TYPE_SIZES[self.protocol.list_size_type]
                    if var.is_simple(True):
                        lines.append(f"size += {accessor}{var.name}.count * {self.protocol.get_size_of(var.vartype)}")
                    elif var.vartype == "string":
                        lines.append(f"for s in {accessor}{var.name} {{")
                        lines.append(f"{self.tab}size += {NUMERIC_TYPE_SIZES[self.protocol.string_size_type]} + s.data(using: String.Encoding.utf8)!.count")
                        lines.append("}")
                    else:
                        lines.append(f"for {var.name}_el in {accessor}{var.name} {{")
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"{var.name}_el.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        clines.append(f"size += {caccum}")
                        lines += [f"{self.tab}{l}" for l in clines]
                        lines.append("}")
                else:
                    if var.is_simple():
                        accum += self.protocol.get_size_of(var.vartype)
                    elif var.vartype == "string":
                        accum += NUMERIC_TYPE_SIZES[self.protocol.string_size_type]
                        lines.append(f"size += {accessor}{var.name}.data(using: String.Encoding.utf8)!.count")
                    else:
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"{accessor}{var.name}.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        lines += clines
                        accum += caccum
        return lines, accum

    def gen_struct(self, sname: str, sdata: Struct):
        if sdata.is_message:
            if self.protocol.namespace != None:
                self.write_line(f"public struct {sname} : {self.protocol.namespace}_Message {{")
            else:
                self.write_line(f"public struct {sname} : Message {{")
        else:
            self.write_line(f"public struct {sname} {{")
        self.indent_level += 1

        for var in sdata.members:
            if var.is_list:
                self.write_line(f"public var {var.name}: [{self.type_mapping[var.vartype]}] = []")
            else:
                default_value = self.base_defaults.get(var.vartype, None)
                if default_value == None:
                    default_value = f"{var.vartype}()"
                self.write_line(f"public var {var.name}: {self.type_mapping[var.vartype]}{f' = {default_value}' if default_value else ''}")
        self.write_line()

        self.write_line("public init() {}")
        self.write_line()

        if sdata.is_message:
            self.write_line("public func GetMessageType() -> MessageType {")
            self.indent_level += 1
            self.write_line(f"return MessageType.{sname}Type")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line()

            self.write_line("public func GetSizeInBytes() -> UInt32 {")
            self.indent_level += 1
            measure_lines, accumulator = self.gen_measurement(sdata, "self.")
            [self.write_line(s) for s in measure_lines]
            if accumulator > 0:
                self.write_line(f"size += {accumulator};")
            if len(measure_lines) > 1:
                self.write_line(f"return UInt32(size)")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line()

            self.write_line(f"public static func FromBytes(_ fromData: Data) -> {sname}? {{")
            self.indent_level += 1
            self.write_line("let dr = DataReader(fromData: fromData)")
            self.write_line("return FromBytes(dataReader: dr)")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line()
            self.write_line(f"static func FromBytes(dataReader: DataReader) -> {sname}? {{")
            self.indent_level += 1
            if len(sdata.members) > 0:
                self.write_line("do {")
                self.indent_level += 1
        else:
            self.write_line(f"static func FromBytes(dataReader: DataReader) throws -> {sname} {{")
            self.indent_level += 1
        decl = "var"
        if len(sdata.members) == 0:
            decl = "let"
        self.write_line(f"{decl} n{sname} = {self.type_mapping[sname]}()")
        [self.deserializer(mem, f"n{sname}.") for mem in sdata.members]
        self.write_line(f"return n{sname}")
        if sdata.is_message and len(sdata.members) > 0:
            self.indent_level -= 1
            self.write_line("}")
            self.write_line("catch {")
            self.indent_level += 1
            self.write_line("return nil")
            self.indent_level -= 1
            self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")

        if sdata.is_message:
            self.write_line("public func WriteBytes(data: inout Data, tag: Bool) -> Void {")
            self.indent_level += 1
            self.write_line("let dataWriter = DataWriter(withData: &data)")
            self.write_line("if (tag) {")
            self.indent_level += 1
            self.write_line(f"dataWriter.WriteUInt8(MessageType.{sname}Type.rawValue)")
            self.indent_level -= 1
            self.write_line("}")
        else:
            self.write_line("func WriteBytes(_ dataWriter: DataWriter) -> Void {")
            self.indent_level += 1
        [self.serializer(mem, "self.") for mem in sdata.members]
        if sdata.is_message:
            self.write_line()
            self.write_line("data = dataWriter.data")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

    def generate(self) -> str:
        self.output = []

        self.write_line(f"// This file was automatically generated by {LIB_NAME} v{LIB_VERSION}")
        self.write_line( "// <https://github.com/sjml/beschi>")
        self.write_line(f"// Do not edit directly.")
        self.write_line()
        self.write_line("import Foundation")
        self.write_line()

        if self.embed_protocol:
            self.write_line("/*")
            self.write_line("DATA PROTOCOL")
            self.write_line("-----------------")
            [self.write_line(f"{l}") for l in self.protocol.protocol_string.splitlines()]
            self.write_line("-----------------")
            self.write_line("END DATA PROTOCOL")
            self.write_line("*/")
            self.write_line()
            self.write_line()

        if self.protocol.namespace != None:
            self.write_line(f"public protocol {self.protocol.namespace}_Message {{")
            self.indent_level += 1
            self.write_line(f"func GetMessageType() -> {self.protocol.namespace}.MessageType")
        else:
            self.write_line("public protocol Message {")
            self.indent_level += 1
            self.write_line(f"func GetMessageType() -> MessageType")
        self.write_line("func WriteBytes(data: inout Data, tag: Bool) -> Void")
        self.write_line("func GetSizeInBytes() -> UInt32")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        if self.protocol.namespace != None:
            self.write_line(f"public /* namespace */ enum {self.protocol.namespace} {{")
            self.indent_level += 1

        self.add_boilerplate(substitutions=[
            ("{# STRING_SIZE_TYPE #}", self.get_native_string_size())
        ])

        self.write_line("public enum MessageType: UInt8 {")
        self.indent_level += 1
        if len(self.protocol.messages) == 0:
            self.write_line("case __NullMessage = 0 /* to keep the compiler happy */")
        [self.write_line(f"case {k}Type = {i+1}") for i, k in enumerate(self.protocol.messages)]
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        if self.protocol.namespace != None:
            self.write_line(f"public static func ProcessRawBytes(_ data: Data) -> [{self.protocol.namespace}_Message?] {{")
            self.indent_level += 1
            self.write_line(f"var msgList: [{self.protocol.namespace}_Message?] = []")
        else:
            self.write_line(f"public func ProcessRawBytes(_ data: Data) -> [Message?] {{")
            self.indent_level += 1
            self.write_line("var msgList: [Message?] = []")
        self.write_line("let dr = DataReader(fromData: data)")
        self.write_line("while !dr.IsFinished() {")
        self.indent_level += 1
        self.write_line("do {")
        self.indent_level += 1
        accessor = ""
        if self.protocol.namespace != None:
            accessor = f"{self.protocol.namespace}."
        self.write_line(f"let msgTypeByte = try dr.GetUInt8()")
        self.write_line(f"guard let msgType = {accessor}MessageType(rawValue: msgTypeByte)")
        self.write_line("else {")
        self.indent_level += 1
        self.write_line("throw DataReader.DataReaderError.InvalidData")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("switch msgType {")
        self.indent_level += 1
        for msg_type in self.protocol.messages:
            self.write_line(f"case {accessor}MessageType.{msg_type}Type:")
            self.indent_level += 1
            self.write_line(f"msgList.append({msg_type}.FromBytes(dataReader: dr))")
            self.indent_level -= 1
        if len(self.protocol.messages) == 0:
            self.write_line(f"case {accessor}MessageType.__NullMessage:")
            self.indent_level += 1
            self.write_line(f"break")
            self.indent_level -= 1
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("catch {")
        self.indent_level += 1
        self.write_line("msgList.append(nil)")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("if msgList.last! == nil {")
        self.indent_level += 1
        self.write_line("break")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("return msgList")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        for sname, sdata in self.protocol.structs.items():
            self.gen_struct(sname, sdata)

        for mname, mdata in self.protocol.messages.items():
            self.gen_struct(mname, mdata)

        if self.protocol.namespace != None:
            self.indent_level -= 1
            self.write_line("}")

        self.write_line()
        assert self.indent_level == 0

        return "\n".join(self.output)



import argparse

from ..protocol import Protocol, Variable, Struct, NUMERIC_TYPE_SIZES
from ..writer import Writer, TextUtil
from .. import LIB_NAME, LIB_VERSION

LANGUAGE_NAME = "TypeScript"


class TypeScriptWriter(Writer):
    language_name = LANGUAGE_NAME
    default_extension = ".ts"

    def get_additional_args(parser: argparse.ArgumentParser):
        group = parser.add_argument_group(LANGUAGE_NAME)
        group.add_argument("--typescript-use-namespace", action="store_const", const=True, default=False, help="wrap generated TypeScript code in a namespace")

    def __init__(self, p: Protocol, extra_args: dict[str,any] = {}):
        super().__init__(protocol=p, tab="  ")

        self.embed_protocol = extra_args["embed_protocol"]

        self.use_namespace = extra_args["typescript_use_namespace"]

        for var_type in NUMERIC_TYPE_SIZES:
            if var_type == "bool":
                self.type_mapping[var_type] = "boolean"
            elif var_type in ["int64", "uint64"]:
                self.type_mapping[var_type] = "bigint"
            else:
                # <sigh>, JavaScript
                self.type_mapping[var_type] = "number"

        self.base_serializers: dict[str,str] = {
            "byte": "Byte",
            "bool": "Bool",
            "uint16": "Uint16",
            "int16": "Int16",
            "uint32": "Uint32",
            "int32": "Int32",
            "uint64": "Uint64",
            "int64": "Int64",
            "float": "Float32",
            "double": "Float64",
        }

    def deserializer(self, var: Variable, accessor: str):
        var_clean = TextUtil.replace(var.name, [("[", "_"), ("]", "_")])
        if var.is_list:
            self.write_line(f"const {var_clean}_Length = da.Get{self.base_serializers[self.protocol.list_size_type]}();")
            self.write_line(f"{accessor}{var.name} = Array<{self.type_mapping[var.vartype]}>({var_clean}_Length);")
            idx = self.indent_level
            self.write_line(f"for (let i{idx} = 0; i{idx} < {var_clean}_Length; i{idx}++) {{")
            self.indent_level += 1
            inner = Variable(self.protocol, f"{var.name}[i{idx}]", var.vartype)
            self.deserializer(inner, accessor)
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype == "string":
            self.write_line(f"{accessor}{var.name} = da.GetString();")
        elif var.vartype in NUMERIC_TYPE_SIZES:
            self.write_line(f"{accessor}{var.name} = da.Get{self.base_serializers[var.vartype]}();")
        else:
            self.write_line(f"{accessor}{var.name} = {var.vartype}.FromBytes(da);")

    def serializer(self, var: Variable, accessor: str):
        if var.is_list:
            self.write_line(f"da.Set{self.base_serializers[self.protocol.list_size_type]}({accessor}{var.name}.length);")
            self.write_line(f"for (let i = 0; i < {accessor}{var.name}.length; i++) {{")
            self.indent_level += 1
            self.write_line(f"let el = {accessor}{var.name}[i];")
            inner = Variable(self.protocol, "el", var.vartype)
            self.serializer(inner, "")
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype == "string":
            self.write_line(f"da.SetString({accessor}{var.name});")
        elif var.vartype in NUMERIC_TYPE_SIZES:
            self.write_line(f"da.Set{self.base_serializers[var.vartype]}({accessor}{var.name});")
        else:
            self.write_line(f"{accessor}{var.name}.WriteBytes(da);")

    def gen_measurement(self, st: Struct, accessor: str = "") -> tuple[list[str], int]:
        lines: list[str] = []
        accum = 0

        if st.is_simple():
            lines.append(f"return {self.protocol.get_size_of(st.name)};")
        else:
            size_init = "let size: number = 0;"
            lines.append(size_init)

            for var in st.members:
                if var.is_list:
                    accum += NUMERIC_TYPE_SIZES[self.protocol.list_size_type]
                    if var.is_simple(True):
                        lines.append(f"size += {accessor}{var.name}.length * {self.protocol.get_size_of(var.vartype)};")
                    elif var.vartype == "string":
                        lines.append(f"for (let {var.name}_i=0; {var.name}_i < {accessor}{var.name}.length; {var.name}_i++) {{")
                        lines.append(f"{self.tab}size += {NUMERIC_TYPE_SIZES[self.protocol.string_size_type]} + _textEnc.encode({accessor}{var.name}[{var.name}_i]).byteLength;")
                        lines.append("}")
                    else:
                        var_clean = TextUtil.replace(var.name, [("[", "_"), ("]", "_")])
                        lines.append(f"for (let {var_clean}_i=0; {var_clean}_i < {accessor}{var.name}.length; {var_clean}_i++) {{")
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"{accessor}{var.name}[{var_clean}_i].")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        clines.append(f"size += {caccum};")
                        lines += [f"{self.tab}{l}" for l in clines]
                        lines.append("}")
                else:
                    if var.is_simple():
                        accum += self.protocol.get_size_of(var.vartype)
                    elif var.vartype == "string":
                        accum += NUMERIC_TYPE_SIZES[self.protocol.string_size_type]
                        lines.append(f"size += _textEnc.encode({accessor}{var.name}).byteLength;")
                    else:
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"{accessor}{var.name}.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        lines += clines
                        accum += caccum
        return lines, accum

    def gen_struct(self, sname: str, sdata: Struct):
        if sdata.is_message:
            self.write_line("@staticImplements<MessageStatic>()")
            self.write_line(f"export class {sname} implements Message {{")
        else:
            self.write_line(f"export class {sname} {{")
        self.indent_level += 1

        for var in sdata.members:
            if var.is_list:
                self.write_line(f"{var.name}: {self.type_mapping[var.vartype]}[] = [];")
            else:
                default_value = "0"
                if self.type_mapping[var.vartype] == "bigint":
                    default_value = "0n"
                elif self.type_mapping[var.vartype] == "boolean":
                    default_value = "false"
                elif self.type_mapping[var.vartype] == "string":
                    default_value = '""'
                elif var.vartype in self.protocol.structs:
                    default_value = f"new {var.vartype}()"
                self.write_line(f"{var.name}: {self.type_mapping[var.vartype]}{f' = {default_value}' if default_value else ''};")
        self.write_line()

        if sdata.is_message:
            self.write_line(f"GetMessageType() : MessageType {{ return MessageType.{sname}Type; }}")
            self.write_line()

            self.write_line("GetSizeInBytes(): number {")
            self.indent_level += 1
            measure_lines, accumulator = self.gen_measurement(sdata, "this.")
            [self.write_line(s) for s in measure_lines]
            if accumulator > 0:
                self.write_line(f"size += {accumulator};")
            if len(measure_lines) > 1:
                self.write_line(f"return size;")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line()

        if sdata.is_message:
            self.write_line(f"static FromBytes(data: DataView|DataAccess): {sname} {{")
            self.indent_level += 1
            self.write_line("let da: DataAccess = null;")
            self.write_line("if (data instanceof DataView) {")
            self.indent_level += 1
            self.write_line("da = new DataAccess(data);")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line("else {")
            self.indent_level += 1
            self.write_line("da = data;")
            self.indent_level -= 1
            self.write_line("}")
        else:
            self.write_line(f"static FromBytes(da: DataAccess): {sname} {{")
            self.indent_level += 1
        if sdata.is_message:
            self.write_line("try {")
            self.indent_level += 1
        self.write_line(f"const n{sname} = new {self.type_mapping[sname]}();")
        [self.deserializer(mem, f"n{sname}.") for mem in sdata.members]
        self.write_line(f"return n{sname};")
        if sdata.is_message:
            self.indent_level -= 1
            self.write_line("}")
            self.write_line("catch (RangeError) {")
            self.indent_level += 1
            self.write_line("return null;")
            self.indent_level -= 1
            self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        if sdata.is_message:
            self.write_line("WriteBytes(data: DataView|DataAccess, tag: boolean): void {")
            self.indent_level += 1
            self.write_line("let da: DataAccess = null;")
            self.write_line("if (data instanceof DataView) {")
            self.indent_level += 1
            self.write_line("da = new DataAccess(data);")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line("else {")
            self.indent_level += 1
            self.write_line("da = data;")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line("if (tag) {")
            self.indent_level += 1
            self.write_line(f"da.SetByte(MessageType.{sname}Type);")
            self.indent_level -= 1
            self.write_line("}")
        else:
            self.write_line("WriteBytes(da: DataAccess) {")
            self.indent_level += 1
        [self.serializer(mem, "this.") for mem in sdata.members]
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

    def generate(self) -> str:
        self.output = []

        self.write_line(f"// This file was automatically generated by {LIB_NAME} v{LIB_VERSION}")
        self.write_line( "// <https://github.com/sjml/beschi>")
        self.write_line(f"// Do not edit directly.")
        self.write_line()

        if self.embed_protocol:
            self.write_line("/*")
            self.write_line("DATA PROTOCOL")
            self.write_line("-----------------")
            [self.write_line(f"{l}") for l in self.protocol.protocol_string.splitlines()]
            self.write_line("-----------------")
            self.write_line("END DATA PROTOCOL")
            self.write_line("*/")
            self.write_line()
            self.write_line()

        if self.use_namespace:
            self.write_line(f"namespace {self.protocol.namespace} {{")
            self.indent_level += 1

        self.add_boilerplate(substitutions=[
            ("{# STRING_SIZE_TYPE #}", self.base_serializers[self.protocol.string_size_type])
        ])

        self.write_line("export enum MessageType {")
        self.indent_level += 1
        [self.write_line(f"{k}Type = {i+1},") for i, k in enumerate(self.protocol.messages)]
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line("export interface Message {")
        self.indent_level += 1
        self.write_line("GetMessageType(): MessageType;")
        self.write_line("WriteBytes(dv: DataView, tag: boolean): void;")
        self.write_line("GetSizeInBytes(): number;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("export interface MessageStatic {")
        self.indent_level += 1
        self.write_line("new(): Message;")
        self.write_line("FromBytes(dv: DataView): Message | null;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("function staticImplements<T>() {")
        self.indent_level += 1
        self.write_line("return (constructor: T) => {}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line("export function ProcessRawBytes(dv: DataView): Message[] {")
        self.indent_level += 1
        self.write_line("const da = new DataAccess(dv);")
        self.write_line("const msgList: Message[] = [];")
        self.write_line("while (!da.IsFinished()) {")
        self.indent_level += 1
        self.write_line("const msgType: number = da.GetByte();")
        self.write_line("switch (msgType) {")
        self.indent_level += 1
        for msg_type in self.protocol.messages:
            self.write_line(f"case MessageType.{msg_type}Type:")
            self.indent_level += 1
            self.write_line(f"msgList.push({msg_type}.FromBytes(da));")
            self.write_line("break;")
            self.indent_level -= 1
        self.write_line("default:")
        self.indent_level += 1
        self.write_line("msgList.push(null);")
        self.write_line("break;")
        self.indent_level -= 1
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("if (msgList[msgList.length - 1] == null) {")
        self.indent_level += 1
        self.write_line("break;")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("return msgList;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        for sname, sdata in self.protocol.structs.items():
            self.gen_struct(sname, sdata)

        for mname, mdata in self.protocol.messages.items():
            self.gen_struct(mname, mdata)

        if self.use_namespace:
            self.indent_level -= 1
            self.write_line("}")

        self.write_line()
        assert self.indent_level == 0

        return "\n".join(self.output)



import os
import sys
import subprocess
import filecmp
import glob

from beschi.writer import Writer
import beschi.writers

CODE_OUTPUT_DIR = "out/generated"
DATA_OUTPUT_DIR = "out/data"
HARNESS_SRC_DIR = "test/_harnesses"
HARNESS_BIN_DIR = "out/executables"

ALL_WRITERS = beschi.writers.all_writers | beschi.writers.experimental_writers
STABLE_WRITERS = beschi.writers.all_writers
EXPERIMENTAL_WRITERS = beschi.writers.experimental_writers

def generate_for(protocol: str, output_name: str, label: str):
    writer_class = ALL_WRITERS[label]

    out_file_dir =  os.path.join(CODE_OUTPUT_DIR, label)
    out_file_path = os.path.join(out_file_dir, f"{output_name}{writer_class.default_extension}")

    if os.path.exists(out_file_path):
        os.unlink(out_file_path)
    assert(not os.path.exists(out_file_path))

    if not os.path.exists(out_file_dir):
        os.makedirs(out_file_dir)

    try:
        # using the CLI instead of calling directly to make sure everything is wired up
        subprocess.check_call(["beschi",
            "--lang", label,
            "--protocol", protocol,
            "--output", out_file_path
        ])
        assert(os.path.exists(out_file_path))
    except Exception as e:
        if label in beschi.writers.experimental_writers:
            sys.stderr.write(f"Experimental writer {label} failed while trying to produce {output_name}.\n")
            sys.stderr.write(f"Exception:\n{e.with_traceback(None)}")
        else:
            raise e

def build_for(language: str, srcfile: str, libfiles: list[str]):
    try:
        writer_class = beschi.writers.all_writers[language]
    except KeyError:
        writer_class = beschi.writers.experimental_writers[language]
    if not os.path.exists(HARNESS_BIN_DIR):
        os.makedirs(HARNESS_BIN_DIR)
    harness_path = os.path.join(HARNESS_SRC_DIR, language)
    libfile_flags = []
    for lf in libfiles:
        libfile_flags.append("--libfile")
        libfile_flags.append(f"{lf}{writer_class.default_extension}")
    invoke_build = [
        "python", f"{harness_path}/builder.py",
        "--srcfile", f"{srcfile}{writer_class.default_extension}",
        *libfile_flags
    ]
    subprocess.check_call(invoke_build + ["--clean"])
    subprocess.check_call(invoke_build)

def run_for(language: str, srcfile: str):
    if not os.path.exists(DATA_OUTPUT_DIR):
        os.makedirs(DATA_OUTPUT_DIR)

    out_file = os.path.join(DATA_OUTPUT_DIR, f"{srcfile}.{language}.msg")
    if os.path.exists(out_file):
        os.unlink(out_file)
    assert(not os.path.exists(out_file))
    subprocess.check_call([
        os.path.join(HARNESS_BIN_DIR, f"{srcfile}_{language}"),
        "--generate", out_file
    ])
    assert(os.path.exists(out_file))

    subprocess.check_call([
        os.path.join(HARNESS_BIN_DIR, f"{srcfile}_{language}"),
        "--read", out_file
    ])

    # HACKHACK
    if language == "c":
        run_for("cpp", srcfile)

def check_files_identical(data_glob):
    filecmp.clear_cache()
    messages = (glob.glob(os.path.join(DATA_OUTPUT_DIR, data_glob)))
    for i in range(len(messages)):
        j = i + 1
        if j >= len(messages):
            j -= len(messages)
        assert(filecmp.cmp(messages[i], messages[j], False))



import os
import re
import argparse

from .protocol import Protocol

DEFAULT_INDENT = "    "

class TextUtil:
    def convert_to_lower_snake_case(s: str) -> str:
        s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s).lower()

    def capitalize(s: str) -> str:
        return s[:1].upper() + s[1:]

    def replace(s: str, subs: list[tuple[str,str]]) -> str:
        for sub in subs:
            s = s.replace(sub[0], sub[1])
        return s


class Writer:
    language_name = "[Base]"
    default_extension = ".beschi"
    in_progress = False # if set to True, writer will not be added to
                        #  the writers.all_writers object, so will be
                        #  shielded from the test suite

    def get_additional_args(parser: argparse.ArgumentParser):
        pass

    def __init__(self, protocol: Protocol, tab: str = DEFAULT_INDENT):
        self.indent_level: int = 0
        self.output: list[str] = []

        self.protocol = protocol
        self.tab: str = tab

        self.type_mapping: dict[str, str] = {}
        self.type_mapping["string"] = "string"
        for struct_type in self.protocol.structs:
            self.type_mapping[struct_type] = struct_type
        for msg_type in self.protocol.messages:
            self.type_mapping[msg_type] = msg_type

    # inserts any boilerplate code, indented at the current level
    def add_boilerplate(self, substitutions: list[tuple[str,str]] = [], index: int =-1):
        if index >= 0:
            index_str = f".{index}"
        else:
            index_str = ""
        filename = f"{self.language_name}{index_str}{self.default_extension}"
        boilerplate_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "writers", "boilerplate", filename)
        if os.path.exists(boilerplate_path):
            boilerplate_lines = open(boilerplate_path, "r").read().splitlines()
            for s in substitutions:
                boilerplate_lines = [bpl.replace(s[0], s[1]) for bpl in boilerplate_lines]
            [self.write_line(bpl) for bpl in boilerplate_lines]
        else:
            print("no file", filename)

    # actually generate the code
    def generate(self) -> str:
        raise NotImplementedError(f"ERROR: 'generate' function needs to be implemented on {self.__class__.__name__}")

    # write line with current indentation level
    def write_line(self, text: str = ""):
        if len(text) == 0:
            self.output.append("")
        else:
            self.output.append(self.indent_string(text))

    # return text with current indentation level
    def indent_string(self, text: str) -> str:
        return (self.tab * self.indent_level) + text

    def get_native_list_size(self) -> str:
        return self.type_mapping[self.protocol.list_size_type]

    def get_native_string_size(self) -> str:
        return self.type_mapping[self.protocol.string_size_type]
